"""Handlers for the reflex_ag_grid component."""

import datetime
from typing import Any, Type, TypeVar

from sqlmodel import not_, and_, or_
from sqlmodel.sql.expression import SelectOfScalar
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.sql.roles import WhereHavingRole

import reflex as rx

M = TypeVar("M", bound=rx.Model)


def handle_text_filter(value, filter_def) -> bool:
    type = filter_def.get("type", "contains")
    filter = filter_def.get("filter", "")
    if type == "contains":
        return filter in value
    if type == "notContains":
        return filter not in value
    if type == "equals":
        return value == filter
    if type == "notEqual":
        return value != filter
    if type == "startsWith":
        return value.startswith(filter)
    if type == "endsWith":
        return value.endswith(filter)
    if type == "blank":
        return not value
    if type == "notBlank":
        return bool(value)
    assert False, f"type {type} does not exist"


def handle_number_filter(value, filter_def) -> bool:
    type = filter_def.get("type", "equals")
    filter = filter_def.get("filter")
    if type == "equals":
        return value == filter
    if type == "notEqual":
        return value != filter
    if type == "greaterThan":
        return value > filter
    if type == "greaterThanOrEqual":
        return value >= filter
    if type == "lessThan":
        return value < filter
    if type == "lessThanOrEqual":
        return value <= filter
    if type == "inRange":
        return filter <= value <= filter_def.get("filterTo")
    if type == "blank":
        return not value
    if type == "notBlank":
        return bool(value)

    assert False, f"type {type} does not exist"


def handle_filter_def(value, filter_def) -> bool:
    if not filter_def:
        return True
    operator = filter_def.get("operator", "").lower()
    if operator == "and":
        return all(
            handle_filter_def(value, sub_filter)
            for sub_filter in filter_def.get("conditions", [])
        )
    elif operator == "or":
        return any(
            handle_filter_def(value, sub_filter)
            for sub_filter in filter_def.get("conditions", [])
        )
    filter_type = filter_def.get("filterType", "text")
    if filter_type == "text":
        return handle_text_filter(value, filter_def)
    if filter_type == "number":
        return handle_number_filter(value, filter_def)
    return False


def handle_filter_model(row, filter_model) -> bool:
    if not filter_model:
        return True
    for field, filter_def in filter_model.items():
        try:
            if not handle_filter_def(row[field], filter_def):
                return False
        except Exception as e:
            print(f"Error filtering {field} of {row}: {e}")
            return False
    return True


_sql_operations = {
    "and": and_,
    "or": or_,
}


def where_text_filter(
    value: InstrumentedAttribute, filter_def: dict[str, str]
) -> WhereHavingRole:
    type = filter_def.get("type", "contains")
    filter = filter_def.get("filter", "")
    if type == "contains":
        return value.contains(filter)
    if type == "notContains":
        return not_(value.contains(filter))
    if type == "equals":
        return value == filter
    if type == "notEqual":
        return value != filter
    if type == "startsWith":
        return value.startswith(filter)
    if type == "endsWith":
        return value.endswith(filter)
    if type == "blank":
        return or_(value == None, value == "")  # noqa: E711
    if type == "notBlank":
        return and_(value != None, value != "")  # noqa: E711
    assert False, f"type {type} does not exist"


def where_number_filter(
    value: InstrumentedAttribute, filter_def: dict[str, str | int | float]
) -> WhereHavingRole:
    type = filter_def.get("type", "equals")
    to_filter = None
    if filter_def.get("filterType") == "date":
        filter = datetime.datetime.fromisoformat(filter_def.get("dateFrom"))
        if filter_def.get("dateTo"):
            to_filter = datetime.datetime.fromisoformat(filter_def.get("dateTo"))
    else:
        filter = filter_def.get("filter", 0)
        if filter_def.get("filterTo"):
            to_filter = filter_def.get("filterTo")
    if type == "equals":
        return value == filter
    if type == "notEqual":
        return value != filter
    if type == "greaterThan":
        return value > filter
    if type == "greaterThanOrEqual":
        return value >= filter
    if type == "lessThan":
        return value < filter
    if type == "lessThanOrEqual":
        return value <= filter
    if type == "inRange":
        return and_(
            value >= filter,
            value <= to_filter,
        )
    if type == "blank":
        return value == None  # noqa: E711
    if type == "notBlank":
        return value != None  # noqa: E711

    assert False, f"type {type} does not exist"


def where_filter_def(
    value: InstrumentedAttribute, filter_def: dict[str, Any]
) -> WhereHavingRole | None:
    if not filter_def:
        return
    operator = _sql_operations.get(
        filter_def.get("operator", "").lower(),
    )
    if operator:
        return operator(
            *(
                where_filter_def(value, sub_filter)
                for sub_filter in filter_def.get("conditions", [])
            )
        )
    filter_type = filter_def.get("filterType", "text")
    if filter_type == "text":
        return where_text_filter(value, filter_def)
    if filter_type in ("number", "date"):
        return where_number_filter(value, filter_def)


def apply_filter_model(
    model: Type[M], filter_model: dict[str, dict[str, Any]]
) -> SelectOfScalar[M]:
    query = model.select()
    for field, filter_def in filter_model.items():
        filter_applies = where_filter_def(
            value=getattr(model, field),
            filter_def=filter_def,
        )
        if filter_applies is not None:
            query = query.where(filter_applies)
    return query


def apply_sort_model(
    model: Type[M], query: SelectOfScalar[M], sort_model: list[dict[str, str]]
) -> SelectOfScalar[M]:
    for sort_spec in sort_model:
        field = getattr(model, sort_spec["colId"], None)
        if field is None:
            continue
        query = query.order_by(
            field.desc() if sort_spec["sort"] == "desc" else field.asc()
        )
    return query



import reflex as rx

from reflex.base import Base
from reflex.config import get_config
from reflex.utils import format
from reflex.utils.imports import ImportVar
from reflex.utils.serializers import serialize


class Datasource(Base):
    uri: str | None = None
    endpoint_uri: str | None = None
    rowCount: rx.Var[int] | int | None = None
    getRows: rx.Var | None = None

    def get_uri(self) -> str:
        return self.uri or (
            f"{self.endpoint_uri}?"
            "start=${params.startRow}"
            "&end=${params.endRow}"
            "&sort_model=${encodeURIComponent(JSON.stringify(params.sortModel))}"
            "&filter_model=${encodeURIComponent(JSON.stringify(params.filterModel))}"
        )

    def _get_rows_function(self) -> str | rx.Var:
        uri = f"getBackendURL(`{get_config().api_url}{self.get_uri()}`)"
        js_func = (
            """
(params) => {
    fetch(%s, {
        headers: {"X-Reflex-Client-Token": token},
    })
    .then((response) => response.json()
        .then((data) => params.successCallback(
            data,
            data.length < params.endRow - params.startRow ? params.startRow + data.length : -1,
        ))
    )
    .catch((error) => params.failCallback())
}
"""
            % uri
        )
        return rx.Var.create_safe(
            js_func.replace("\n", ""),
            _var_is_local=False,
            _var_is_string=False,
            _var_data=rx.vars.VarData(
                imports={
                    "/utils/state": [
                        ImportVar(tag="getBackendURL"),
                        ImportVar(tag="token"),
                    ],
                }
            ),
        )

    def dict(self, **kwargs):
        d = super().dict(**kwargs)
        d.pop("uri", None)
        if self.getRows is None:
            d["getRows"] = self._get_rows_function()
        return d

    def json(self) -> str:
        """Convert the object to a json-like string.

        Vars will be unwrapped so they can represent actual JS var names and functions.

        Keys will be converted to camelCase.

        Returns:
            The object as a Javascript Object literal.
        """
        d = self.dict(exclude={"uri"})
        if self.getRows is None:
            d["getRows"] = self._get_rows_function()
        return format.unwrap_vars(
            self.__config__.json_dumps(
                {format.to_camel_case(key): value for key, value in d.items()},
                default=serialize,
            )
        )


class SSRMDatasource(Base):
    uri: str | None = None
    endpoint_uri: str | None = None
    getRows: rx.Var | None = None

    def get_uri(self) -> str:
        return self.uri or (
            f"{self.endpoint_uri}?"
            "start=${params.request.startRow}"
            "&end=${params.request.endRow}"
            "&rowGroupCols=${encodeURIComponent(JSON.stringify(params.request.rowGroupCols))}"
            "&groupKeys=${encodeURIComponent(JSON.stringify(params.request.groupKeys))}"
            "&valueCols=${encodeURIComponent(JSON.stringify(params.request.valueCols))}"
            "&pivotMode=${params.request.pivotMode}"
            "&sortModel=${encodeURIComponent(JSON.stringify(params.request.sortModel))}"
            "&filterModel=${encodeURIComponent(JSON.stringify(params.request.filterModel))}"
            "&pivotCols=${encodeURIComponent(JSON.stringify(params.request.pivotCols))}"
        )

    def _get_rows_function(self) -> str | rx.Var:
        uri = f"getBackendURL(`{get_config().api_url}{self.get_uri()}`)"
        js_func = (
            """
(params) => {
    fetch(%s, {
        headers: {"X-Reflex-Client-Token": token},
    })
    .then((response) => response.json()
        .then((data) => {
            params.success(data)
        })
    )
    .catch((error) => params.fail())
}
"""
            % uri
        )
        return rx.Var.create_safe(
            js_func.replace("\n", ""),
            _var_is_local=False,
            _var_is_string=False,
            _var_data=rx.vars.VarData(
                imports={
                    "/utils/state": [
                        ImportVar(tag="getBackendURL"),
                        ImportVar(tag="token"),
                    ],
                }
            ),
        )

    def json(self) -> str:
        """Convert the object to a json-like string.

        Vars will be unwrapped so they can represent actual JS var names and functions.

        Keys will be converted to camelCase.

        Returns:
            The object as a Javascript Object literal.
        """
        d = self.dict(exclude={"uri"})
        if self.getRows is None:
            d["getRows"] = self._get_rows_function()
        return format.unwrap_vars(
            self.__config__.json_dumps(
                {format.to_camel_case(key): value for key, value in d.items()},
                default=serialize,
            )
        )



"""ag-grid rx.model wrapper."""

from __future__ import annotations

import datetime
import enum
import json
from typing import Any, ClassVar, Generic, Type

from fastapi import Request
from sqlmodel import col, func, select

import reflex as rx

from reflex_ag_grid.ag_grid import ag_grid, ColumnDef
from reflex_ag_grid.datasource import Datasource
from reflex_ag_grid.handlers import apply_filter_model, apply_sort_model, M


def _value_setter_signature(
    params: rx.Var[dict[str, Any]],
) -> tuple[rx.Var[int], rx.Var[str], rx.Var[Any]]:
    return (
        params["data"],
        params["colDef"].to(dict)["field"],
        params["newValue"],
    )


def get_default_column_def(
    field: str, ftype: Type, value_setter: rx.EventHandler | None = None, **cdef_kwargs
) -> ColumnDef:
    """Get a default column definition for a given field type.

    Args:
        field: The field name.
        ftype: The field type.
        value_setter: The value setter event handler.
        **cdef_kwargs: Additional ColumnDef keyword arguments.

    Returns:
        The column definition.
    """
    _cdef_kwargs = dict(
        sortable=True,
        filter=True,
        editable=True if value_setter is not None else False,
        cell_editor=ag_grid.editors.text,
    )
    _cdef_kwargs.update(cdef_kwargs)
    cdef = ag_grid.column_def(
        field=field,
        **_cdef_kwargs,
    )
    if rx.utils.types.is_optional(ftype):
        ftype = rx.utils.types.args(ftype)[0]
    if ftype in (int, float, bool):
        cdef.type = "numericColumn"
        cdef.filter = ag_grid.filters.number
        cdef.cell_editor = ag_grid.editors.number
    if ftype is bool:
        cdef.cell_renderer = "agCheckboxCellRenderer"
        cdef.cell_editor = ag_grid.editors.checkbox
    if ftype is datetime.datetime:
        cdef.filter = ag_grid.filters.date
        cdef.cell_editor = ag_grid.editors.date
    if value_setter is not None:
        cdef.value_setter = rx.EventChain(
            events=[
                rx.event.call_event_handler(value_setter, _value_setter_signature),
            ],
            args_spec=_value_setter_signature,
            # XXX: hacks to queue events from call_script eval context
            invocation=rx.vars.function.FunctionStringVar(
                "((events) => {queueEvents(events, socket); processEvent(socket)})",
            ),
        )
    return cdef


class AbstractWrapper(rx.ComponentState):
    """Abstract class for wrapping ag-grid for infinite data model."""

    _grid_component: ClassVar[rx.Component | None] = None
    _selected_items: list[Any] = []

    __data_route__ = "/abstract-wrapper-data"
    __get_data_kwargs__ = {
        "state": lambda cls: cls.get_full_name(),
        "start": "${params.startRow}",
        "end": "${params.endRow}",
        "sort_model": "${encodeURIComponent(JSON.stringify(params.sortModel))}",
        "filter_model": "${encodeURIComponent(JSON.stringify(params.filterModel))}",
    }

    @classmethod
    def _add_data_route(cls):
        """Add the backend __data_route__ that responds to ag-grid data requests.

        The backend route will call the _get_data method to fetch the data.
        """
        app = rx.utils.prerequisites.get_app().app
        if cls.__data_route__ in app.api.routes:
            return

        @app.api.get(cls.__data_route__)
        async def get_data(
            request: Request,
            state: str,
            start: int,
            end: int,
            filter_model: str = None,
            sort_model: str = None,
        ):
            try:
                token = request.headers["X-Reflex-Client-Token"]
            except KeyError:
                return []
            if filter_model is not None:
                filter_model = json.loads(filter_model)
            if sort_model is not None:
                sort_model = json.loads(sort_model)
            state_cls = rx.State.get_class_substate(tuple(state.split(".")))
            async with app.modify_state(
                rx.state._substate_key(token, state_cls)
            ) as state:
                s_instance = await state.get_state(state_cls)
                result = s_instance._get_data(
                    start=start,
                    end=end,
                    filter_model=filter_model,
                    sort_model=sort_model,
                )
                if hasattr(result, "__await__"):
                    result = await result
                return result

    @classmethod
    def _get_datasource_uri(cls) -> str:
        """Get the uri for the ag-grid DataSource model."""
        return (
            cls.__data_route__
            + "?"
            + "&".join(
                f"{key}={value if not callable(value) else value(cls)}"
                for key, value in cls.__get_data_kwargs__.items()
            )
        )

    def on_mount(self):
        """Perform post-hydration grid initialization.

        Set up column defs and data source to fetch infinite row data.
        """
        return [
            self._grid_component.api.set_grid_option(
                "columnDefs", self._get_column_defs()
            ),
            self._grid_component.set_datasource(
                Datasource(
                    uri=self._get_datasource_uri(),
                    rowCount=self._row_count(),
                ),
            ),
        ]

    def on_selection_changed(self, rows, source, type):
        """Handle selection changed event."""
        self._selected_items = rows

    def on_value_setter(self, row_data: dict[str, Any], field_name: str, value: Any):
        """Handle setting value in the model."""
        raise NotImplementedError("Handle setting value in the model.")

    def _get_column_defs(self) -> list[ColumnDef]:
        """Get the column definitions for the grid, must be overridden."""
        raise NotImplementedError("Handle fetching column definitions.")

    def _get_data(
        self,
        start: int,
        end: int,
        filter_model: dict[str, Any] | None = None,
        sort_model: list[dict[str, str]] | None = None,
    ) -> list[dict[str, Any]]:
        """Get the data for the grid, must be overridden."""
        raise NotImplementedError("Handle fetching data from the model.")

    def _row_count(self) -> int:
        """Get the total row count for the grid, must be overridden."""
        raise NotImplementedError("Handle fetching row count.")

    @classmethod
    def get_component(cls, *children, **props) -> rx.Component:
        """Return the Ag-Grid component linked to the wrapper state.

        Args:
            children: The children components passed to ag_grid, typically not used.
            **props: Additional props for the ag_grid component.

        Note that "column_defs", "row_model_type", "on_mount", and
        "on_selection_change" cannot be passed here.

        Returns:
            The Ag-Grid component.
        """
        _props = dict(
            id=f"ag-grid-{cls.get_full_name()}",
            default_col_def={"flex": 1},
            max_blocks_in_cache=4,
            cache_block_size=50,
            group_default_expanded=None,
        )
        _props.update(props)
        return ag_grid.root(
            *children,
            row_model_type="infinite",
            on_mount=cls.on_mount,
            on_selection_changed=cls.on_selection_changed,
            **_props,
        )

    @classmethod
    def create(cls, *children, **props) -> rx.Component:
        """Create the ComponentState instance.

        Args:
            children: The children components passed to ag_grid, typically not used.
            **props: Additional props for the ag_grid component.

        Note that "column_defs", "row_model_type", "on_mount", and
        "on_selection_change" cannot be passed here.

        Returns:
            The Ag-Grid component linked to the wrapper state.
        """
        comp = super().create(*children, **props)
        comp.State._grid_component = comp
        comp.State._add_data_route()
        return comp


class ModelWrapperActionType(enum.Enum):
    """ModelWrapper action types."""

    SELECT = "select"
    INSERT = "insert"
    UPDATE = "update"
    DELETE = "delete"


class ModelWrapper(AbstractWrapper, Generic[M]):
    """Ag-Grid wrapper for arbitrary rx.Model class."""

    _model_class: ClassVar[Type[M] | None] = None
    _selected_items: list[M] = []
    add_dialog_is_open: bool = False

    async def _is_authorized(
        self,
        action: ModelWrapperActionType,
        action_data: list[M] | dict[str, Any] | None,
    ) -> bool:
        """Check if the user is authorized to perform the action.

        For SELECT, action_data is None.
        For INSERT, action_data is a dictionary of the new row data.
        For UPDATE, action data is a dictionary of updated row data.
        For DELETE, action data is a list of model objects to delete.

        Args:
            action: The action type.
            action_data: The data associated with the action.

        Returns:
            True if authorized, False otherwise.
        """
        return True

    def on_selection_changed(self, rows, source, type):
        self._selected_items = [self._model_class(**row) for row in rows]

    async def on_value_setter(
        self, row_data: dict[str, Any], field_name: str, value: Any
    ):
        if not await self._is_authorized(
            ModelWrapperActionType.UPDATE, row_data | {field_name: value}
        ):
            return
        try:
            if self._model_class.__fields__[field_name].type_ == datetime.datetime:
                value = datetime.datetime.fromisoformat(value)
        except KeyError:
            pass
        with rx.session() as session:
            item_orm = session.get(self._model_class, row_data["id"])
            if item_orm is not None:
                setattr(item_orm, field_name, value)
                session.add(item_orm)
                session.commit()
                return self._grid_component.api.refreshInfiniteCache()

    async def on_add(self, row_data: dict[str, Any]):
        """Handles submitting a new row to the model."""
        if not await self._is_authorized(ModelWrapperActionType.INSERT, row_data):
            return
        with rx.session() as session:
            item = self._model_class(**row_data)
            session.add(item)
            session.commit()
            self.add_dialog_is_open = False
            return self._grid_component.api.refreshInfiniteCache()

    async def delete_selected(self):
        """Handles deleting selected rows from the model."""
        if not await self._is_authorized(
            ModelWrapperActionType.DELETE, self._selected_items
        ):
            return
        with rx.session() as session:
            for item in session.exec(
                self._model_class.select().where(
                    self._model_class.id.in_([item.id for item in self._selected_items])
                )
            ).all():
                session.delete(item)
            session.commit()
            return self._grid_component.api.refreshInfiniteCache()

    def _get_column_defs(self) -> list[ColumnDef]:
        return [
            get_default_column_def(
                field=field.name,
                ftype=field.type_,
                value_setter=type(self).on_value_setter,
                editable=True if field.name != "id" else False,
            )
            for field in self._model_class.__fields__.values()
        ]

    def _row_count(self) -> int:
        with rx.session() as session:
            return session.exec(select(func.count(col(self._model_class.id)))).one()

    async def _get_data(
        self,
        start: int,
        end: int,
        filter_model: dict[str, Any] | None = None,
        sort_model: list[dict[str, str]] | None = None,
    ) -> list[M]:
        if not await self._is_authorized(ModelWrapperActionType.SELECT, None):
            return []
        with rx.session() as session:
            return session.exec(
                apply_sort_model(
                    model=self._model_class,
                    query=apply_filter_model(
                        model=self._model_class,
                        filter_model=filter_model or {},
                    ),
                    sort_model=sort_model or [],
                )
                .offset(start)
                .limit(end - start)
            ).all()

    @classmethod
    def _add_dialog_field(cls, field: str, ftype: Type) -> rx.Component:
        comp = rx.input(name=field)
        orig_type_str = str(ftype)
        if rx.utils.types.is_optional(ftype):
            ftype = rx.utils.types.args(ftype)[0]
        if ftype in (int, float):
            comp = rx.input(
                name=field,
                type="number",
            )
        if ftype is bool:
            comp = rx.checkbox(id=field)
        return rx.table.row(
            rx.table.cell(rx.text(field)),
            rx.table.cell(comp),
            rx.table.cell(rx.text(f"({orig_type_str})")),
        )

    @classmethod
    def _add_dialog(cls) -> rx.Component:
        """Create the dialog for adding a new row."""
        return rx.dialog.root(
            rx.dialog.trigger(
                rx.icon_button("plus"),
            ),
            rx.dialog.content(
                rx.dialog.title(f"Add new {cls._model_class.__name__}"),
                rx.form(
                    rx.table.root(
                        *[
                            cls._add_dialog_field(field.name, field.type_)
                            for field in cls._model_class.__fields__.values()
                            if field.name != "id"
                        ]
                    ),
                    rx.button("Add", margin="5px"),
                    on_submit=cls.on_add,
                ),
            ),
            open=cls.add_dialog_is_open,
            on_open_change=cls.set_add_dialog_is_open,
        )

    @classmethod
    def _delete_button(cls) -> rx.Component:
        """Create the delete button."""
        return rx.icon_button(
            "minus",
            on_click=cls.delete_selected,
        )

    @classmethod
    def _top_toolbar(cls) -> rx.Component:
        """Create the top toolbar."""
        return rx.hstack(
            cls._delete_button(),
            cls._add_dialog(),
            justify="end",
            margin="5px",
        )

    @classmethod
    def create(cls, *children, model_class: Type[M], **props) -> rx.Component:
        comp = super().create(*children, **props)
        comp.State._model_class = model_class
        return rx.fragment(
            comp.State._top_toolbar(),
            comp,
        )


model_wrapper = ModelWrapper.create



"""Reflex custom component AgGrid."""

# For wrapping react guide, visit https://reflex.dev/docs/wrapping-react/overview/

from types import SimpleNamespace
import reflex as rx

import os
from typing import Any, Callable, Union
from .datasource import Datasource, SSRMDatasource
from reflex.components.props import PropsBase
from typing import Literal
from reflex.components.el import Div


def callback_content(iterable: list[str]) -> str:
    return "; ".join(iterable)


def arrow_callback(js_expr: str | list[str]):
    if isinstance(js_expr, list):
        js_expr = callback_content(js_expr)
    return rx.Var(f"(() => {{{js_expr}}})()")


def exclude_non_serializable_keys(
    event: rx.Var,
    exclude_keys: list[str],
    log_event: bool = False,
) -> list[str]:
    exclude_keys_str = ", ".join(exclude_keys)

    exprs = [
        f"let {{{exclude_keys_str}, ...rest}} = {event}",
        "return rest",
    ]

    if log_event:
        exprs = [f"console.log({event})", *exprs]
    return exprs


def _on_cell_event_spec(event: rx.Var) -> list[rx.Var]:
    # Remove non-serializable keys from the event object
    exclude_keys = [
        "context",
        "api",
        "columnApi",
        "column",
        "node",
        "event",
        "eventPath",
    ]
    return [
        arrow_callback(
            exclude_non_serializable_keys(
                event,
                exclude_keys,
                True,
            )
        ),
    ]


def _on_row_event_spec(event: rx.Var) -> list[rx.Var]:
    exclude_keys = ["context", "api", "source", "node", "event", "eventPath"]
    return [
        arrow_callback(
            exclude_non_serializable_keys(
                event,
                exclude_keys,
                True,
            )
        ),
    ]


def _on_column_event_spec(event: rx.Var) -> list[rx.Var]:
    return [
        arrow_callback(
            [
                f"console.log({event})",
                f"let {{type, column, colDef, api, ...rest}} = {event}",
                "let columnID = column.colId",
                "return {type, columnID}",
            ]
        ),  # ID of the column being clicked
    ]


def _on_row_selected(event: rx.Var) -> list[rx.Var]:
    return [
        arrow_callback(
            [
                f"let {{type, node, data, rowIndex, ...rest}} = {event}",
                "return {type, data, rowIndex}",
            ]
        )
    ]


def _on_cell_value_changed(event: rx.Var) -> list[rx.Var]:
    return [
        rx.Var(
            f"(() => {{let {{rowIndex, ...rest}} = {event}; console.log({event}) ; return rowIndex}})()"
        ),  # index of the row being changed
        rx.Var(
            f"(() => {{let {{colDef, ...rest}} = {event}; return colDef.field}})()"
        ),  # field of the column being changed
        rx.Var(
            f"(() => {{let {{newValue, ...rest}} = {event}; return newValue}})()"
        ),  # new value
    ]


def _on_selection_change_signature(event: rx.Var) -> list[rx.Var]:
    return [
        rx.Var(f"{event}.api.getSelectedRows()"),
        rx.Var(f"{event}.source"),
        rx.Var(f"{event}.type"),
    ]


size_columns_to_fit = rx.Var(
    "(event) => event.api.sizeColumnsToFit()", _var_type=rx.EventChain
)


class AGFilters(SimpleNamespace):
    text = "agTextColumnFilter"
    number = "agNumberColumnFilter"
    date = "agDateColumnFilter"
    set = "agSetColumnFilter"
    multi = "agMultiColumnFilter"


class AGEditors(SimpleNamespace):
    text = "agTextCellEditor"
    large_text = "agLargeTextCellEditor"
    select = "agSelectCellEditor"
    rich_select = "agRichSelectCellEditor"
    number = "agNumberCellEditor"
    date = "agDateCellEditor"
    checkbox = "agCheckboxCellEditor"


class AGRenderers(SimpleNamespace):
    link = rx.vars.function.ArgsFunctionOperation.create(
        ("params",),
        rx.link(
            rx.Var("params.value"),
            href=rx.Var("params.value", _var_type=str),
            target="_blank",
        ),
    ).to(dict)  # TODO: remove cast after reflex 0.6.6


class ColumnDef(PropsBase):
    field: str | rx.Var[str]
    col_id: str | rx.Var[str] | None = None
    type: str | rx.Var[str] | None = None
    cell_data_type: bool | str | rx.Var[bool] | rx.Var[str] | None = None
    hide: bool | rx.Var[bool] = False
    editable: bool | rx.Var[bool] = False
    filter: AGFilters | str | rx.Var[AGFilters] | rx.Var[str] | None = None
    floating_filter: bool | rx.Var[bool] = False
    header_name: str | rx.Var[str] | None = None
    header_tooltip: str | rx.Var[str] | None = None
    checkbox_selection: bool | rx.Var[bool] = False
    cell_editor: AGEditors | str | rx.Var[AGEditors] | rx.Var[str] | None = None
    cell_editor_params: dict[str, list[Any]] | rx.Var[dict[str, list[Any]]] | None = (
        None
    )
    value_setter: rx.EventChain | rx.Var[rx.EventChain] | None = None
    value_formatter: rx.Var | None = None
    wrap_text: bool | None = None
    auto_height: bool | None = None
    auto_header_height: bool | None = None
    enable_cell_change_flash: bool | None = None
    cell_editor_popup: bool | None = None
    cell_editor_popup_position: str | None = None
    resizable: bool | None = None
    suppress_span_header_height: bool | None = None
    cell_renderer: rx.Var | None = None
    flex: int | rx.Var[int] | None = None


class ColumnGroup(PropsBase):
    children: list["ColumnDef | ColumnGroup"] | rx.Var[list["ColumnDef | ColumnGroup"]]
    group_id: str | rx.Var[str]
    marry_children: bool | rx.Var[bool] = False
    open_by_default: bool | rx.Var[bool] = False
    column_group_show: Literal["open", "closed"] | rx.Var[str] = "open"
    header_name: str | rx.Var[str]
    header_tooltip: str | rx.Var[str] | None = None


class AgGridAPI(rx.Base):
    ref: str

    @classmethod
    def create(cls, id: str) -> "AgGridAPI":
        """Create an instance of the AgGridAPI class.

        Args:
            id: The ID of the AgGrid component.

        Returns:
            An instance of the AgGridAPI class.
        """
        return cls(ref=rx.utils.format.format_ref(id))

    @property
    def _api(self) -> rx.Var:
        return f"refs['{self.ref}']?.current?.api"

    def __getattr__(self, name: str) -> Callable[..., rx.event.EventSpec]:
        def _call_api(*args, **kwargs) -> rx.event.EventSpec:
            """Call the ag-grid API method with the given arguments.

            Args:
                *args: Arguments to pass to the API method.
                **kwargs: Keyword arguments to pass to rx.call_script.

            Returns:
                rx.event.EventSpec: The event specification.
            """
            var_args = [str(rx.Var.create(arg)) for arg in args]
            return rx.call_script(
                f"{self._api}.{rx.utils.format.to_camel_case(name)}({', '.join(var_args)})",
                **kwargs,
            )

        return _call_api


class AgGrid(rx.Component):
    """Reflex AgGrid component is a high-performance and highly customizable component that wraps AG Grid, designed for creating rich datagrids."""

    # The library name for the ag-grid-react component
    library: str = "ag-grid-react@32.1.0"

    # The tag name for the AgGridReact component
    tag: str = "AgGridReact"

    # Variable for column definitions
    column_defs: rx.Var[list[dict[str, Any] | ColumnDef | ColumnGroup]]

    # Variable for row data
    row_data: rx.Var[list[dict[str, Any]]]

    # Variable for cell selection
    cell_selection: bool | rx.Var[bool] = False

    # Variable for row selection type
    row_selection: rx.Var[str] = "single"

    # Variable to animate rows
    animate_rows: rx.Var[bool] = False

    # Variable for pagination
    pagination: rx.Var[bool] = False

    # Page size for pagination
    pagination_page_size: rx.Var[int] = rx.Var.create(10)

    # Strategy for auto sizing
    auto_size_strategy: rx.Var[dict] = rx.Var.create({})

    # Selector for pagination page size options
    pagination_page_size_selector: rx.Var[list[int]] = rx.Var.create([10, 25, 50])

    # Variable for the side bar configuration
    side_bar: rx.Var[Union[str, dict[str, Any], bool, list[str]]] = rx.Var.create("")

    # Variable to indicate if tree data is used
    tree_data: rx.Var[bool] = rx.Var.create(False)

    # Default column definition
    default_col_def: rx.Var[dict[str, Any]] = rx.Var.create({})

    # Definition for the auto group column
    auto_group_column_def: rx.Var[Any] = rx.Var.create({})

    # Data for pinned top rows
    pinned_top_row_data: rx.Var[list[dict[str, Any]]] = rx.Var.create([])

    # Data for pinned bottom rows
    pinned_bottom_row_data: rx.Var[list[dict[str, Any]]] = rx.Var.create([])

    # Default expanded group level
    group_default_expanded: rx.Var[int] | None = rx.Var.create(-1)

    # Variable to indicate if group selects children
    group_selects_children: rx.Var[bool] = rx.Var.create(False)

    # Variable to suppress row click selection
    suppress_row_click_selection: rx.Var[bool] = rx.Var.create(False)

    # Event handler for getting the data path
    get_data_path: rx.EventHandler[lambda e0: [e0]]

    # Variable to allow unbalanced groups
    group_allow_unbalanced: rx.Var[bool] = rx.Var.create(False)

    # Variable to show pivot panel
    pivot_panel_show: rx.Var[str] = rx.Var.create("never")

    # Variable to show row group panel
    row_group_panel_show: rx.Var[str] = rx.Var.create("never")

    # Variable to suppress aggregate function in header
    suppress_agg_func_in_header: rx.Var[bool] = rx.Var.create(False)

    # Variable to lock group columns
    group_lock_group_columns: rx.Var[int] = rx.Var.create(0)

    # Variable to maintain column order
    maintain_column_order: rx.Var[bool] = rx.Var.create(False)

    # Row model type for infinite/serverside row model
    row_model_type: rx.Var[str]

    # Cache block size for infinite/serverside row model
    cache_block_size: rx.Var[int]

    # Maximum blocks in cache for infinite/serverside row model
    max_blocks_in_cache: rx.Var[int]

    # Row buffer size for infinite/serverside row model
    row_buffer: rx.Var[int]

    # Cache overflow size for infinite/serverside row model
    cache_overflow_size: rx.Var[int]

    # Maximum concurrent datasource requests for infinite/serverside row model
    max_concurrent_datasource_requests: rx.Var[int]

    # Initial row count for infinite row model
    infinite_initial_row_count: rx.Var[int]

    # Datasource for infinite/serverside row model
    datasource: rx.Var[Datasource]

    # Event handler for getting the row ID
    get_row_id: rx.EventHandler[lambda e0: [e0]]

    # Event handler to check if it is a server-side group
    is_server_side_group: rx.EventHandler[lambda e0: [e0]]

    # Event handler to get the server-side group key
    get_server_side_group_key: rx.EventHandler[lambda e0: [e0]]

    # Server-side datasource for infinite/serverside row model
    server_side_datasource: rx.Var[SSRMDatasource]

    # Event handler to check if server-side group is open by default
    is_server_side_group_open_by_default: rx.EventHandler[lambda e0: [e0]]

    # Variable to enable client-side sort on server-side
    server_side_enable_client_side_sort: rx.Var[bool] = False

    # Event handler to get the child count
    get_child_count: rx.EventHandler[lambda e0: [e0]]

    # Event handler for cell click events
    on_cell_clicked: rx.EventHandler[_on_cell_event_spec]

    # Event handler for cell focused events
    on_cell_focused: rx.EventHandler[_on_cell_event_spec]

    # Event handler for cell mouse over events
    on_cell_mouse_over: rx.EventHandler[_on_cell_event_spec]

    # Event handler for cell mouse out events
    on_cell_mouse_out: rx.EventHandler[_on_cell_event_spec]

    # Event handler for cell double click events
    on_cell_double_clicked: rx.EventHandler[_on_cell_event_spec]

    # Event handler for right click on a cell
    on_cell_context_menu: rx.EventHandler[_on_cell_event_spec]

    # Event handler for row data changed events
    on_cell_value_changed: rx.EventHandler[_on_cell_value_changed]

    # Event handler for row click events
    on_row_clicked: rx.EventHandler[_on_row_event_spec]

    # Event handler for row double click events
    on_row_double_clicked: rx.EventHandler[_on_row_event_spec]

    # Event handler for row selected events
    on_row_selected: rx.EventHandler[_on_row_selected]

    # Event handler for selection change events
    on_selection_changed: rx.EventHandler[_on_selection_change_signature]

    # Event handler for column header clicked events
    on_column_header_clicked: rx.EventHandler[_on_column_event_spec]

    # Event handler for column resized events
    on_column_resized: rx.EventHandler[_on_column_event_spec]

    # Event handler for column moved events
    on_column_moved: rx.EventHandler[_on_column_event_spec]

    # Event handler for column pinned events
    on_column_pinned: rx.EventHandler[_on_column_event_spec]

    # Event handler for column header context menu events
    on_column_header_context_menu: rx.EventHandler[_on_column_event_spec]

    # Event handler for column header focused events
    on_header_focused: rx.EventHandler[_on_column_event_spec]

    # Event handler for first data rendered events
    on_first_data_rendered: rx.EventHandler[_on_cell_event_spec]

    lib_dependencies: list[str] = [
        "ag-grid-community@32.1.0",
        "ag-grid-enterprise@32.1.0",
    ]
    # Change the aesthetic theme of the grid
    theme: rx.Var[Literal["quartz", "balham", "alpine", "material"]]

    # Event handler for when the grid is ready
    on_grid_ready: rx.EventHandler[lambda e0: [e0]]

    @classmethod
    def create(
        cls,
        *children,
        id: str,
        data_path_key: str | None = None,
        is_server_side_group_key: str | None = None,
        get_server_side_group_key: str | None = None,
        server_side_group_open_level: int | None = None,
        child_count_key: str | None = None,
        row_id_key: str | None = None,
        **props,
    ) -> rx.Component:
        props.setdefault("id", id)

        # handle hierarchical data
        if data_path_key is not None:
            props["get_data_path"] = rx.Var(f"(data) => data.{data_path_key}").to(
                rx.EventChain
            )

        if is_server_side_group_key is not None:
            props["is_server_side_group"] = rx.Var(
                f"(data) => data.{is_server_side_group_key}"
            ).to(rx.EventChain)

        if get_server_side_group_key is not None:
            props["get_server_side_group_key"] = rx.Var(
                f"(data) => data.{get_server_side_group_key}",
            ).to(rx.EventChain)

        if server_side_group_open_level is not None:
            props["server_side_group_open_level"] = rx.Var(
                f"(params) => params.rowNode.level < {server_side_group_open_level}"
            ).to(rx.EventChain)

        if child_count_key is not None:
            props["get_child_count"] = rx.Var(
                f"(data) => data ? data.{child_count_key} : undefined"
            ).to(rx.EventChain)

        if row_id_key is not None:
            props["get_row_id"] = rx.Var(f"(params) => params.data.{row_id_key}").to(
                rx.EventChain
            )

        props["class_name"] = rx.match(
            props.get("theme", "quartz"),
            ("quartz", rx.color_mode_cond("ag-theme-quartz", "ag-theme-quartz-dark")),
            ("balham", rx.color_mode_cond("ag-theme-balham", "ag-theme-balham-dark")),
            ("alpine", rx.color_mode_cond("ag-theme-alpine", "ag-theme-alpine-dark")),
            (
                "material",
                rx.color_mode_cond("ag-theme-material", "ag-theme-material-dark"),
            ),
            "",
        )

        if "auto_size_strategy" in props:
            props["on_grid_ready"] = size_columns_to_fit

        return super().create(*children, **props)

    def add_imports(self):
        return {
            "": [
                "ag-grid-community/styles/ag-grid.css",
                "ag-grid-community/styles/ag-theme-quartz.css",
                "ag-grid-community/styles/ag-theme-balham.css",
                "ag-grid-community/styles/ag-theme-material.css",
                "ag-grid-community/styles/ag-theme-alpine.css",
                "ag-grid-enterprise",
            ],
            "d3-format": ["format"],
            "ag-grid-enterprise": [
                "LicenseManager",
            ],
        }

    def add_custom_code(self) -> list[str]:
        ag_grid_license_key = os.getenv("AG_GRID_LICENSE_KEY")
        if ag_grid_license_key is not None:
            return [f"LicenseManager.setLicenseKey('{ag_grid_license_key}');"]
        return ["LicenseManager.setLicenseKey(null);"]

    @property
    def api(self) -> AgGridAPI:
        return AgGridAPI(ref=self.get_ref())

    def get_selected_rows(self, callback: rx.EventHandler) -> rx.event.EventSpec:
        return self.api.getSelectedRows(callback=callback)

    def select_all(self) -> rx.event.EventSpec:
        return self.api.selectAll()

    def deselect_all(self) -> rx.event.EventSpec:
        return self.api.deselectAll()

    def select_rows_by_key(self, keys: list[str]) -> rx.event.EventHandler:
        keys_var = rx.Var.create(keys, _var_is_string=False)
        script = f"""
let api = {self.api};
const selected_nodes = [];
let keys_set = new Set({keys_var});
api.forEachNode(function (node) {{
    if (keys_set.has(node.key)) {{
        selected_nodes.push(node);
    }} 
}});
api.deselectAll()
api.setNodesSelected({{ nodes: selected_nodes, newValue: true }});
"""
        return rx.call_script(script)  # type: ignore

    def log_nodes(self) -> rx.event.EventSpec:
        return rx.call_script(
            f"""
let api = {self.api};
console.log("Logging nodes");
api.forEachNode(function (node) {{
    console.log(node.key);
}});
"""
        )

    def setGridOption(self, key: str, value: rx.Var) -> rx.event.EventSpec:
        return self.api.set_grid_option(key, value)

    def set_datasource(self, datasource: Datasource):
        return self.setGridOption(
            key="datasource",
            value=rx.Var.create(datasource),
        )

    def set_serverside_datasource(self, datasource: SSRMDatasource):
        return self.setGridOption(
            key="serverSideDatasource",
            value=rx.Var.create(datasource, _var_is_string=False),
        )

    def show_loading_overlay(self) -> rx.event.EventSpec:
        return self.api.showLoadingOverlay()

    def show_no_rows_overlay(self) -> rx.event.EventSpec:
        return self.api.showNoRowsOverlay()

    def hide_overlay(self) -> rx.event.EventSpec:
        return self.api.hideOverlay()

    def redraw_rows(self) -> rx.event.EventSpec:
        return self.api.redrawRows()

    def export_data_as_csv(self) -> rx.event.EventSpec:
        """Export the grid data as a CSV file."""
        return self.api.exportDataAsCsv()


class WrappedAgGrid(AgGrid):
    """Reflex AgGrid component is a high-performance and highly customizable component that wraps AG Grid, designed for creating rich datagrids."""

    @classmethod
    def create(cls, *children, **props):
        width = props.pop("width", None)
        height = props.pop("height", None)
        return Div.create(
            super().create(*children, **props),
            width=width or "40vw",
            height=height or "25vh",
        )


class AgGridNamespace(rx.ComponentNamespace):
    api = AgGridAPI.create
    column_def = ColumnDef
    column_group = ColumnGroup
    filters = AGFilters
    editors = AGEditors
    renderers = AGRenderers
    size_columns_to_fit = size_columns_to_fit
    root = AgGrid.create
    __call__ = WrappedAgGrid.create


ag_grid = AgGridNamespace()



from .ag_grid import ag_grid
from .datasource import Datasource
from .handlers import (
    apply_filter_model,
    apply_sort_model,
    handle_filter_def,
    handle_filter_model,
    handle_number_filter,
    handle_text_filter,
    where_filter_def,
    where_number_filter,
    where_text_filter,
)
from .wrapper import (
    AbstractWrapper,
    ModelWrapper,
    ModelWrapperActionType,
    model_wrapper,
)

__all__ = [
    "AbstractWrapper",
    "Datasource",
    "ModelWrapper",
    "ModelWrapperActionType",
    "ag_grid",
    "apply_filter_model",
    "apply_sort_model",
    "handle_filter_def",
    "handle_filter_model",
    "handle_number_filter",
    "handle_text_filter",
    "model_wrapper",
    "where_number_filter",
    "where_text_filter",
    "where_filter_def",
]
